--Gameboy Hyperram Bridge

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity Gameboy_Hyperram_Bridge is
    port(
        GB_nAHB_MODE : in std_logic;
        GB_nRST     : in std_logic;
        clk  : in std_logic;
        --Gamboy Interface
        nWR     : in std_logic;
        nRD     : in std_logic;
        nCS     : in std_logic;
        A       : in std_logic_vector(14 downto 0);
        A15     : in std_logic;
        D       : inout std_logic_vector(7 downto 0);
        DIR     : out std_logic;
        nOE     : out std_logic;
        
        --HYPERRAM
		GB_wr_data: out std_logic_vector(31 downto 0);
		GB_rd_data: in std_logic_vector(31 downto 0);
		GB_rd_data_valid: in std_logic;
		GB_addr: out std_logic_vector(21 downto 0);
		GB_cmd: out std_logic;
		GB_cmd_en: out std_logic;
		GB_init_calib: in std_logic;
		GB_clk_out: in std_logic;
        GB_data_mask: out std_logic_vector(3 downto 0)
);
end Gameboy_Hyperram_Bridge;

architecture rtl of Gameboy_Hyperram_Bridge is

type ZUSTAENDE is (IDLE, R1, R2, R3, R4, R5, R6, R7, R8, R9, R10, R11, R12, R13, R14, R15, R16, R17, R18, R19, R20, R21, R22, R23, R24, R25);
signal ZUSTAND, FOLGE_Z: ZUSTAENDE;

signal Din: std_logic_vector (7 downto 0);
signal Dout: std_logic_vector(7 downto 0);

begin

	Z_SPEICHER: process(GB_clk_out, GB_nRST, GB_nAHB_MODE)
	begin
        if GB_nAHB_MODE = '0' or GB_nRST = '0' then 
            ZUSTAND <= IDLE;
        elsif GB_clk_out='1' and GB_clk_out'event then
			ZUSTAND <= FOLGE_Z ;
		end if;
	end process Z_SPEICHER;

	UE_SN: process(ZUSTAND, nCS, nRD, nWR, A15, GB_rd_data_valid)
	begin
		case ZUSTAND is
			when IDLE 	=>	if nRD='0' and A15='0' then 
								FOLGE_Z <= R1 ;
							else
								FOLGE_Z <= IDLE ;
							end if;
			when R1		=>	FOLGE_Z	<= R2 ;
			when R2		=>	FOLGE_Z	<= R3 ;
			when R3		=>	FOLGE_Z	<= R4 ;
			when R4		=>	FOLGE_Z	<= R5 ;
			when R5		=>	FOLGE_Z	<= R6 ;
			when R6		=>	FOLGE_Z	<= R7 ;
			when R7		=>	FOLGE_Z	<= R8 ;
			when R8		=>	FOLGE_Z	<= R9 ;
			when R9		=>	FOLGE_Z	<= R10 ;
			when R10		=>	FOLGE_Z	<= R11 ;
			when R11		=>	FOLGE_Z	<= R12 ;
			when R12		=>	FOLGE_Z	<= R13 ;
			when R13		=>	FOLGE_Z	<= R14 ;
			when R14		=>	FOLGE_Z	<= R15 ;
			when R15		=>	FOLGE_Z	<= R16 ;
			when R16		=>	FOLGE_Z	<= R17 ;
			when R17		=>	FOLGE_Z	<= R18 ;
			when R18		=>	FOLGE_Z	<= R19 ;
			when R19		=>	FOLGE_Z	<= R20 ;
			when R20		=>	FOLGE_Z	<= R21 ;
			when R21	    =>	FOLGE_Z	<= R22 ;
			when R22	    =>	FOLGE_Z	<= R23 ;
			when R23	    =>	FOLGE_Z	<= R24 ;
			when R24	    =>	FOLGE_Z	<= R25 ;
			when R25    	=>	FOLGE_Z	<= IDLE ;
			
			when others	=> 	FOLGE_Z <= IDLE ;
		end case;
	end process UE_SN;
	
	AUS_SN: process(ZUSTAND, A, Din)
	begin
		case ZUSTAND is
			when IDLE	=>	Dout <= x"00";
			when R1	=>	Dout <=     "00000000";
			when R2	=>	Dout <=     "00000000";
			when R3	=>	Dout <=     "00000000";
			when R4	=>	Dout <=     "00000000";
			when R5	=>	Dout <=     "00000000";
			when R6	=>	Dout <=     "00000000";
			when R7	=>	Dout <=     "00000000";
			when R8	=>	Dout <=     "00000000";
			when R9	=>	Dout <=     "00000001";
			when R10	=>	Dout <= "00000011"; --Read
			when R11	=>	Dout <= "00000010"; --Read
			when R12	=>	Dout <= "00000110"; 
			when R13	=>	Dout <= "11111111";
			when R14	=>	Dout <= "11111111";
			when R15	=>	Dout <= "11111111";
			when R16	=>	Dout <= "11111111";
			when R17	=>	Dout <= "11111111";
			when R18	=>	Dout <= "11111111"; --HYPERRAM_rd_valid
			when R19	=>	Dout <= "11111111"; --HYPERRAM Dout
			when R20	=>	Dout <= "11111111";
			when R21	=>	Dout <= "11111111";
			when R22	=>	Dout <= "11111111";
			when R23	=>	Dout <= "11111111";
			when R24	=>	Dout <= "11111111";
			when R25	=>	Dout <= "11111111";

			when others =>	Dout <= x"00";
		end case;
	end process AUS_SN;

--	AUS_SYNC: process(GB_clk_out)
--	begin
--		if GB_clk_out='1' and GB_clk_out'event then
--			if FOLGE_Z = R3 then
--				Dout <= GB_rd_data(23 downto 16) ;
--			end if;
--		end if;
--	end process AUS_SYNC;

--	AUS_SYNC: process(GB_rd_data_valid)
--	begin
--		if GB_rd_data_valid='1' and GB_rd_data_valid'event then
--            Dout <= GB_rd_data(23 downto 16) ;
--		end if;
--	end process AUS_SYNC;
	
    GB_data_mask <= "0000";
 
    GB_wr_data <= (others => '0');
    GB_addr <= (others => '0');
    
    GB_cmd <= '0';
    GB_cmd_en <= '0';

    DATA_OUT: process (nCS, nRD, A15, Dout, nWR)
    begin
        if not nRD and not A15 then
            D <= Dout;
            DIR <= '1';
            nOE <= '0';
--        elsif not nCS and not nWR and not A15 then
--            D <= (others => 'Z');
--            DIR <= '1';
--            nOE <= '1';
        else
--           D <= (others => 'Z');
            D <= Dout;
            DIR <= '1';
            nOE <= '1';
        end if;
    end process;

    Din <= D;

end rtl;

