
--AHB HyperRAM
--Default is Burst 16 -> 4 word

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity Gowin_AHB_HyperRAM_Top is
    port(
        --AHB
        AHB_HRDATA: out std_logic_vector(31 downto 0);
        AHB_HREADY: out std_logic;
        AHB_HRESP: out std_logic;
        AHB_HTRANS: in std_logic_vector(1 downto 0);
        AHB_HBURST: in std_logic_vector(2 downto 0);
        AHB_HPROT: in std_logic_vector(3 downto 0);
        AHB_HSIZE: in std_logic_vector(2 downto 0);
        AHB_HWRITE: in std_logic;
        AHB_HMASTLOCK: in std_logic;
        AHB_HMASTER: in std_logic_vector(3 downto 0);
        AHB_HADDR: in std_logic_vector(31 downto 0);
        AHB_HWDATA: in std_logic_vector(31 downto 0);
        AHB_HSEL: in std_logic;
        AHB_HCLK: in std_logic;
        AHB_HRESETn: in std_logic;

        --Gamboy Interface
        nWR     : in std_logic;
        nRD     : in std_logic;
        nCS     : in std_logic;
        A       : in std_logic_vector(14 downto 0);
        A15     : in std_logic;
        D       : inout std_logic_vector(7 downto 0);
        DIR     : out std_logic;
        nOE     : out std_logic;
        
        --HYPERRAM
        W_HPRAM_WRDATA: out std_logic_vector(31 downto 0);
        hpram_data_out: in std_logic_vector(31 downto 0);
        rd_valid: in std_logic;
        W_reg_address: out std_logic_vector(21 downto 0);
        W_cross_reg_cmd2: out std_logic;
        W_cmd_enable: out std_logic;
        wire_init: in std_logic;
        user_clk: in std_logic;
        data_mask: out std_logic_vector(3 downto 0);

        --GB
        GB_RESET: out std_logic;
        S_GB_nAHB_MODE: out std_logic
);
end Gowin_AHB_HyperRAM_Top;

architecture rtl of Gowin_AHB_HyperRAM_Top is

--Gameboy

signal Din: std_logic_vector (7 downto 0);
signal Dout: std_logic_vector(7 downto 0);

--MBC
signal BANK1: std_logic_vector (4 downto 0);
signal BANK2: std_logic_vector ( 1 downto 0);

signal MBCA: std_logic_vector (21 downto 0);

-- Address change
signal Aold: std_logic_vector(14 downto 0);
signal Aevent: std_logic;
signal Aeventlate: std_logic;

signal maskhold: std_logic_vector( 11 downto 0);
signal mask: std_logic;

-- Program Counter for Debug
signal count: std_logic_vector(15 downto 0);
signal countii: std_logic_vector(7 downto 0);
signal countvalid: std_logic_vector(7 downto 0);
signal countvalidmask: std_logic;


--AHB


--Define Reg for AHB BUS
signal S_ahb_address: std_logic_vector (11 downto 0);


signal S_write_enable: std_logic;
signal S_read_enable: std_logic;

--//Write to AHB register

signal S_reg_cmd: std_logic;
signal S_GB_RESET: std_logic;
signal S_S_GB_nAHB_MODE: std_logic;

signal S_reg_address: std_logic_vector(21 downto 0);
signal S_reg_cmd_en: std_logic;
signal S_reg_data_in0, S_reg_data_in1, S_reg_data_in2, S_reg_data_in3:std_logic_vector(31 downto 0);
signal S_reg_read_done: std_logic;



--//read data from HyperRAM
signal S_reg_data_out0, S_reg_data_out1, S_reg_data_out2, S_reg_data_out3: std_logic_vector(31 downto 0);
signal S_ahb_rdata: std_logic_vector(31 downto 0);

--//wire wire_init;
--//cross the clk domain -init sig
--//user clk -> ahb clk
signal S_cross_wire_init1, S_cross_wire_init2: std_logic;


--//cross the reg_cmd ahb clk -> user clk
signal S_cross_reg_cmd1, S_cross_reg_cmd2: std_logic;


--//cross the reg_cmd_en ahb clk -> user clk
signal S_cross_reg_cmd_en1, S_cross_reg_cmd_en2: std_logic;
signal S_raise_edge_reg_cmd_en: std_logic;


--//the raise edge of the reg cmd enable
signal S_cmd_enable: std_logic;

--//Send data to HyperRAM
--//write data to HyperRAM

signal S_wr_state: std_logic_vector(2 downto 0);


signal S_hpram_wrdata: std_logic_vector(31 downto 0);

--//read data from HyperRAM

signal READ_STATE: std_logic_vector(3 downto 0);
signal READ_STATE_FOLGE: std_logic_vector(3 downto 0);


--//cross the rd valid user clk -> ahb clk
signal S_cross_rd_valid1, S_cross_rd_valid2: std_logic;

--//get the fall edge of rd valid
signal S_fall_edge_rd_valid1, S_fall_edge_rd_valid2: std_logic;
signal S_fall_edge_rd_valid: std_logic;

begin
    --The AHB BUS is always ready
    AHB_HREADY <= '1';
    --Response OKAY
    AHB_HRESP  <= '0';

  


  P1: process(AHB_HCLK, AHB_HRESETn)
    begin
    if AHB_HRESETn = '0' then
        S_ahb_address <= (others => '0');
        S_write_enable <= '0';
        S_read_enable <= '0';
    elsif AHB_HCLK ='1' and AHB_HCLK'event then
        S_ahb_address <= AHB_HADDR (11 downto 0);

        if AHB_HTRANS(1) and AHB_HSEL then
            if AHB_HWRITE then
                S_write_enable <= '1';
                S_read_enable <= '0';
            else
                S_write_enable <= '0';
                S_read_enable <= '1';
            end if;
        else
            S_write_enable <= '0';
            S_read_enable <= '0';
        end if;

    end if;

    end process;


--//write data to register	
    P2: process(AHB_HCLK, AHB_HRESETn)
    begin
    if (AHB_HRESETn ='0') then
        S_reg_cmd <= '0';
        S_GB_RESET <= '1';
        S_S_GB_nAHB_MODE <= '0';
        S_reg_address <= (others => '0');

        S_reg_data_in0 <= (others => '0');
        S_reg_data_in1 <= (others => '0');
        S_reg_data_in2 <= (others => '0');
        S_reg_data_in3 <= (others => '0');
     elsif AHB_HCLK = '1' and AHB_HCLK'event then
        if (S_write_enable) then
            case S_ahb_address(4 downto 2) is
                when "000" =>   S_reg_cmd <= AHB_HWDATA(0);
                                S_GB_RESET <= AHB_HWDATA(1);
                                S_S_GB_nAHB_MODE <= AHB_HWDATA(2);
                when "001" => S_reg_address <= AHB_HWDATA (21 downto 0);
                when "010" => S_reg_data_in0 <= AHB_HWDATA;
                when "011" => S_reg_data_in1 <= AHB_HWDATA;
                when "100" => S_reg_data_in2 <= AHB_HWDATA;
                when "101" => S_reg_data_in3 <= AHB_HWDATA;
                when "110" => S_reg_cmd_en <= AHB_HWDATA(0);
                when others => NULL;
            end case;
        end if;
    end if;
    end process;

--//read data from AHB

    P3: process( S_read_enable, S_ahb_address, S_reg_cmd, S_GB_RESET, S_S_GB_nAHB_MODE, S_reg_address, S_reg_data_in0, S_reg_data_in1, S_reg_data_in2, S_reg_data_in3, 
                S_reg_cmd_en, S_reg_read_done, S_reg_data_out0, S_reg_data_out1, S_reg_data_out2, S_reg_data_out3, S_cross_wire_init2)
    begin
        if(S_read_enable) then
--            case S_ahb_address(11 downto 2) is
--                when "0000000000" => S_ahb_rdata <= "0000000000000000000000000000000" & S_reg_cmd;
--                when "0000000001" => S_ahb_rdata <= "0000000000" & S_reg_address;
--                when "0000000010" => S_ahb_rdata <= S_reg_data_in0;
--                when "0000000011" => S_ahb_rdata <= S_reg_data_in1;
--                when "0000000100" => S_ahb_rdata <= S_reg_data_in2;
--                when "0000000101" => S_ahb_rdata <= S_reg_data_in3;
--                when "0000000110" => S_ahb_rdata <= "0000000000000000000000000000000" & S_reg_cmd_en;

--                when "0000000111" => S_ahb_rdata <= "0000000000000000000000000000000" & S_reg_read_done;
--                when "0000001000" => S_ahb_rdata <= S_reg_data_out0;
--                when "0000001001" => S_ahb_rdata <= S_reg_data_out1;
--                when "0000001010" => S_ahb_rdata <= S_reg_data_out2;
--                when "0000001011" => S_ahb_rdata <= S_reg_data_out3;
--                when "0000001100" => S_ahb_rdata <= "0000000000000000000000000000000" & S_cross_wire_init2;
--    
--                when others => S_ahb_rdata <= (others => '1');

            case S_ahb_address(5 downto 2) is            
                when "0000" => S_ahb_rdata <= "00000000000000000000000000000"& S_S_GB_nAHB_MODE & S_GB_RESET & S_reg_cmd;
                when "0001" => S_ahb_rdata <= "0000000000" & S_reg_address;
                when "0010" => S_ahb_rdata <= S_reg_data_in0;
                when "0011" => S_ahb_rdata <= S_reg_data_in1;
                when "0100" => S_ahb_rdata <= S_reg_data_in2;
                when "0101" => S_ahb_rdata <= S_reg_data_in3;
                when "0110" => S_ahb_rdata <= "0000000000000000000000000000000" & S_reg_cmd_en;

                when "0111" => S_ahb_rdata <= "0000000000000000000000000000000" & S_reg_read_done;
                when "1000" => S_ahb_rdata <= S_reg_data_out0;
                when "1001" => S_ahb_rdata <= S_reg_data_out1;
                when "1010" => S_ahb_rdata <= S_reg_data_out2;
                when "1011" => S_ahb_rdata <= S_reg_data_out3;
                when "1100" => S_ahb_rdata <= "0000000000000000000000000000000" & S_cross_wire_init2;
               
                when others => S_ahb_rdata <= (others => '1');

            end case;
        end if;
    end process;

    AHB_HRDATA <= S_ahb_rdata;


--//cross_wire_init2 will be read

    P4: process(AHB_HRESETn, AHB_HCLK)
    begin
        if(AHB_HRESETn ='0') then
            S_cross_wire_init1 <= '0';
            S_cross_wire_init2 <= '0';
        elsif AHB_HCLK = '1' and AHB_HCLK'event then
            S_cross_wire_init1 <= wire_init;
            S_cross_wire_init2 <= S_cross_wire_init1;
        end if;
    end process;



--//cross_reg_cmd2 will be use
    P5: process(AHB_HRESETn, user_clk)
    begin
        if AHB_HRESETn ='0' then
            S_cross_reg_cmd1 <= '0';
            S_cross_reg_cmd2 <= '0';
        elsif user_clk ='1' and user_clk'event then
            S_cross_reg_cmd1 <= S_reg_cmd;
            S_cross_reg_cmd2 <= S_cross_reg_cmd1;
        end if;
    end process;


    P6: process (user_clk, AHB_HRESETn)
    begin
        if AHB_HRESETn ='0' then
            S_cross_reg_cmd_en1 <= '0';
            S_cross_reg_cmd_en2 <= '0';
        elsif user_clk ='1' and user_clk'event then
            S_cross_reg_cmd_en1 <= S_reg_cmd_en;
            S_cross_reg_cmd_en2 <= S_cross_reg_cmd_en1;
        end if;
    end process;

    S_raise_edge_reg_cmd_en <= S_cross_reg_cmd_en1 and (not S_cross_reg_cmd_en2);

    P7: process(user_clk, AHB_HRESETn)
    begin
        if AHB_HRESETn = '0' then
            S_cmd_enable <= '0';
        elsif user_clk ='1' and user_clk'event then
            if S_raise_edge_reg_cmd_en = '1' then
                S_cmd_enable <= '1';
            else
                S_cmd_enable <= '0';
            end if;
        end if;
    end process;


    P8: process(S_S_GB_nAHB_MODE, AHB_HRESETn, user_clk)
    begin
        if AHB_HRESETn = '0' or S_S_GB_nAHB_MODE = '1' then
            S_wr_state <= "000";
        elsif user_clk='1' and user_clk'event then
  --	begin//write cmd and vist the ahb bus and set the cmd enable function
          case S_wr_state is
                when "000" =>    if S_cross_reg_cmd2 and S_cmd_enable then
                                    S_wr_state <= S_wr_state + 1;
                                end if;
                when "001" =>    S_wr_state <= S_wr_state + 1;
                when "010" =>    S_wr_state <= S_wr_state + 1;
                when "011" =>    S_wr_state <= S_wr_state + 1;
                when "100" =>    S_wr_state <= S_wr_state + 1;
                when "101" =>    S_wr_state <= S_wr_state + 1;
                when "110" =>    S_wr_state <= S_wr_state + 1;
                when "111" =>    S_wr_state <= S_wr_state + 1;
            end case;
        end if;
    end process;

    P9: process(S_wr_state, S_reg_data_in0, S_reg_data_in1, S_reg_data_in2, S_reg_data_in3)
    begin
        case S_wr_state is
            when "000" => S_hpram_wrdata <= S_reg_data_in0;
                            data_mask <= "0000";
            when "001" => S_hpram_wrdata <= S_reg_data_in1;
                            data_mask <= "0000";
            when "010" => S_hpram_wrdata <= S_reg_data_in2;
                            data_mask <= "0000";
            when "011" => S_hpram_wrdata <= S_reg_data_in3;
                            data_mask <= "0000";
            when "100" => S_hpram_wrdata <= (others => '0');
                            data_mask <= "1111";
            when "101" => S_hpram_wrdata <= (others => '0');
                            data_mask <= "1111";
            when "110" => S_hpram_wrdata <= (others => '0');
                            data_mask <= "1111";
            when "111" => S_hpram_wrdata <= (others => '0');
                            data_mask <= "1111";
        end case;
    end process;


    W_HPRAM_WRDATA <= S_hpram_wrdata;

    S_rd_state_SPEICHER: process (user_clk, AHB_HRESETn, S_S_GB_nAHB_MODE)
    begin
         if AHB_HRESETn = '0' then
            READ_STATE <= "0000";
        elsif user_clk='1' and user_clk'event then 
            READ_STATE <= READ_STATE_FOLGE;
        end if;
    end process;


    UE_S_rd_state: process(READ_STATE, S_cross_reg_cmd2, S_cmd_enable, rd_valid, nCS, nRD, nWR, A15, rd_valid, Aevent)
    begin
        case READ_STATE is
            --IDLE
            when "0000" =>  if (not S_cross_reg_cmd2) and S_cmd_enable then
                                READ_STATE_FOLGE <= "0001";
                            elsif Aevent = '1' then
                                READ_STATE_FOLGE <= "1001";
                            else    
                                READ_STATE_FOLGE <= "0000";
                            end if;
            --AHB
            when "0001" =>   if (rd_valid) then
                                READ_STATE_FOLGE <= "0010";
                            else
                                READ_STATE_FOLGE <= "0001";
                            end if;
            when "0010" =>   READ_STATE_FOLGE <= "0011";
            when "0011" =>   READ_STATE_FOLGE <= "0100";
            when "0100" =>   READ_STATE_FOLGE <= "0000";

            --GB
			when "1001"	=>	READ_STATE_FOLGE	<= "1010" ;
			when "1010"	=>	if rd_valid='1' then 
                                READ_STATE_FOLGE <= "1011" ;
                            else
                                READ_STATE_FOLGE <= "1010" ;
                            end if;
			when "1011"	=> READ_STATE_FOLGE <= "1100";
			when "1100"	=>	READ_STATE_FOLGE <= "0000" ;
            when others =>   READ_STATE_FOLGE <= "0000";
        end case;
    end process;

	AUS_SN: process(READ_STATE, A, Din, S_reg_address, S_cross_reg_cmd2, S_cmd_enable, MBCA)
	begin
		case READ_STATE is

			when "0000"	=>	
                            W_reg_address <= S_reg_address;
							W_cross_reg_cmd2 <= S_cross_reg_cmd2;
                            W_cmd_enable <= S_cmd_enable;
			--AHB				
			when "0001"		=>	
                            W_reg_address <= S_reg_address;		-- Read Start
							W_cross_reg_cmd2 <= S_cross_reg_cmd2;
                            W_cmd_enable <= S_cmd_enable;
			when "0010"		=>	
                            W_reg_address <= S_reg_address;
							W_cross_reg_cmd2 <= S_cross_reg_cmd2;
                            W_cmd_enable <= S_cmd_enable;
			when "0011"		=>	
                            W_reg_address <= S_reg_address;
							W_cross_reg_cmd2 <= S_cross_reg_cmd2;
                            W_cmd_enable <= S_cmd_enable;
			when "0100"		=>	
                            W_reg_address <= S_reg_address;
							W_cross_reg_cmd2 <= S_cross_reg_cmd2;
                            W_cmd_enable <= S_cmd_enable;
			--GB				
			when "1001"		=>	
                            W_reg_address <= MBCA;		-- Read Start
							W_cross_reg_cmd2 <= '0';
                            W_cmd_enable <= '1';
			when "1010"		=>	
                            W_reg_address <= (others=>'0');
							W_cross_reg_cmd2 <= '0';
                            W_cmd_enable <= '0';
			when "1011"		=>	
                            W_reg_address <= (others=>'0');
							W_cross_reg_cmd2 <= '0';
                            W_cmd_enable <= '0';
			when "1100"		=>	
                            W_reg_address <= (others=>'0');
							W_cross_reg_cmd2 <= '0';
                            W_cmd_enable <= '0';
			when others =>	
                            W_reg_address <= (others=>'0');
							W_cross_reg_cmd2 <= '0';
                            W_cmd_enable <= '0';
		end case;
	end process AUS_SN;
	

    AUS_S_rd_state: process(user_clk)
    begin
        if user_clk='1' and user_clk'event then 
            case READ_STATE is
                --IDLE
                when "0000" =>   null;
                --AHB
                when "0001" =>   if (rd_valid) then
                                    S_reg_data_out0 <= hpram_data_out;
                                end if;
                when "0010" =>   S_reg_data_out1 <= hpram_data_out;
                when "0011" =>   S_reg_data_out2 <= hpram_data_out;
                when "0100" =>   S_reg_data_out3 <= hpram_data_out;
                --GB
                when "1010" =>  if READ_STATE_FOLGE = "1011" then
                                    Dout <= hpram_data_out(23 downto 16) ;
                                end if;
                when others => null;
            end case;
        end if;
    end process;
   
 
    P11: process (AHB_HCLK, AHB_HRESETn)
    begin
        if AHB_HRESETn = '0' then   
            S_cross_rd_valid1 <= '0';
            S_cross_rd_valid2 <= '0';
        elsif AHB_HCLK='1' and AHB_HCLK'event then
            S_cross_rd_valid1 <= rd_valid;
            S_cross_rd_valid2 <= S_cross_rd_valid1;
        end if;
    end process;


    P12: process (AHB_HCLK, AHB_HRESETn)
    begin
        if AHB_HRESETn = '0' then
            S_fall_edge_rd_valid1 <= '0';
            S_fall_edge_rd_valid2 <= '0';
        elsif AHB_HCLK = '1' and AHB_HCLK'event then
            S_fall_edge_rd_valid1 <= S_cross_rd_valid2;
            S_fall_edge_rd_valid2 <= S_fall_edge_rd_valid1;
        end if;
    end process;

    S_fall_edge_rd_valid <= (not S_fall_edge_rd_valid1) and S_fall_edge_rd_valid2;


    P13: process(AHB_HCLK, AHB_HRESETn)
    begin
        if AHB_HRESETn = '0' then
            S_reg_read_done <= '0';
        elsif AHB_HCLK = '1' and AHB_HCLK'event then
            if (S_write_enable='1') and (S_ahb_address(11 downto 2) = "0000000111") then
                S_reg_read_done <= AHB_HWDATA(0);
            elsif S_fall_edge_rd_valid then
                S_reg_read_done <= '1';
            end if;
        end if;
    end process;
 

--    W_reg_address <= S_reg_address;
--    W_cross_reg_cmd2 <= S_cross_reg_cmd2;
--    W_cmd_enable <= S_cmd_enable;


    GB_RESET <= S_GB_RESET;
    S_GB_nAHB_MODE <= S_S_GB_nAHB_MODE;

    --Gameboy

--    data_mask <= "0000";

    --MBC
   MBCA_SN: process (BANK1, BANK2, A)
    begin
        if A(14) ='0' then
            MBCA <= "00000000" & A(13 downto 0);
        else        --A(14) ='1' 
            if BANK2="00" and BANK1="00000" then     --BANK1 doesn't map 00000
                MBCA <= "00000001" & A(13 downto 0);
            else
                MBCA <= '0' & BANK2 & BANK1 & A(13 downto 0);
            end if;
        end if;
    end process MBCA_SN;

    BANK_SPEICHER: process(nWR, S_GB_RESET, S_GB_nAHB_MODE, A15, A, Din, BANK1, BANK2)
	begin
        if S_GB_nAHB_MODE = '0' or S_GB_RESET = '0' then 
            BANK1 <= "00001";
            BANK2 <= "00";
        elsif nWR = '0' then
			if A15 = '0' and A(14) = '0' and A(13) = '1' then
                BANK1 <= Din(4 downto 0);
                BANK2 <= BANK2;
            elsif A15 = '0' and A(14) = '1' and A(13) = '0' then
                BANK1 <= BANK1;
                BANK2 <= Din(1 downto 0);
            else
                BANK1 <= BANK1;
                BANK2 <= BANK2;
            end if;
		end if;
	end process BANK_SPEICHER;
--    
--   maskhold_SPEICHER: process(user_clk, S_GB_RESET, S_GB_nAHB_MODE)
--	begin
--        if S_GB_nAHB_MODE = '0' or S_GB_RESET = '0' then 
--            maskhold <= (others=> '0');
--        elsif user_clk='1' and user_clk'event then
--			if A15 = '0' then
--                maskhold <= (others => '0');
--            else    
--                maskhold <= maskhold + 1;
--            end if;
--		end if;
--	end process maskhold_SPEICHER;

--    hold_SPEICHER: process(user_clk, S_GB_RESET, S_GB_nAHB_MODE)
--	begin
--        if S_GB_nAHB_MODE = '0' or S_GB_RESET = '0' then 
--            mask <= '0';
--        elsif user_clk='1' and user_clk'event then
--            if    maskhold = x"000" or maskhold = x"04b" or maskhold = x"096" or maskhold = x"0e2" or maskhold = x"12e" then
--                mask <= '0';
--			elsif maskhold = x"008" or maskhold = x"052" or maskhold = x"09e" or maskhold = x"0e7" or maskhold = x"134" then
--                mask <= '1';
--            else    
--                mask <= mask;
--            end if;
--		end if;
--	end process hold_SPEICHER;  

   maskhold_SPEICHER: process(user_clk, S_GB_RESET, S_GB_nAHB_MODE)
	begin
        if S_GB_nAHB_MODE = '0' or S_GB_RESET = '0' then 
            maskhold <= (others=> '0');
        elsif user_clk='1' and user_clk'event then
			if Aevent = '1' then
                maskhold <= (others => '0');
            else    
                maskhold <= maskhold + 1;
            end if;
		end if;
	end process maskhold_SPEICHER;

    hold_SPEICHER: process(user_clk, S_GB_RESET, S_GB_nAHB_MODE)
	begin
        if S_GB_nAHB_MODE = '0' or S_GB_RESET = '0' then 
            mask <= '0';
        elsif user_clk='1' and user_clk'event then
            if                          maskhold = x"039" or maskhold = x"085" or maskhold = x"0d1" or maskhold = x"11d" then
                mask <= '0';
			elsif maskhold = x"000" or maskhold = x"043" or maskhold = x"08f" or maskhold = x"0db" or maskhold = x"127" then
                mask <= '1';
            else    
                mask <= mask;
            end if;
		end if;
	end process hold_SPEICHER;  

    Aevent_SPEICHER: process(user_clk, S_GB_RESET, S_GB_nAHB_MODE)
	begin
        if S_GB_nAHB_MODE = '0' or S_GB_RESET = '0' then 
            Aevent <= '0';
        elsif user_clk='1' and user_clk'event then
			if A = Aold or mask = '0' then
                Aevent <= '0';
            else    
                Aevent <= '1';
            end if;
		end if;
	end process Aevent_SPEICHER;

--   Aeventlate_SPEICHER: process(user_clk, S_GB_RESET, S_GB_nAHB_MODE)
--	begin
--        if S_GB_nAHB_MODE = '0' or S_GB_RESET = '0' then 
--            Aeventlate <= '0';
--        elsif user_clk='1' and user_clk'event then
--            if READ_STATE = "1010" then
--                Aeventlate <= '1';
--            else    
--                Aeventlate <= '0';
--            end if;
--		end if;
--	end process Aeventlate_SPEICHER;

   Aold_SPEICHER: process(user_clk, S_GB_RESET, S_GB_nAHB_MODE)
	begin
        if S_GB_nAHB_MODE = '0' or S_GB_RESET = '0' then 
            Aold <= (others => '0');
        elsif user_clk='1' and user_clk'event then
			if Aevent = '1' or mask = '0' then
                Aold <= A;
            else    
                Aold <= Aold;
            end if;

		end if;
	end process Aold_SPEICHER;

--   count_SPEICHER: process(A15, S_GB_RESET, S_GB_nAHB_MODE)
--	begin
--        if S_GB_nAHB_MODE = '0' or S_GB_RESET = '0' then 
--            count <= (others => '0');
--        elsif A15='1' and A15'event then
--			count <= count + 1;
--		end if;
--	end process count_SPEICHER;

--  countii_SPEICHER: process(count(15), S_GB_RESET, S_GB_nAHB_MODE)
--	begin
--        if S_GB_nAHB_MODE = '0' or S_GB_RESET = '0' then 
--            countii <= (others => '0');
--        elsif count(15)='0' and count(15)'event then
--			countii <= countii + 1;
--		end if;
--	end process countii_SPEICHER;


--    countvalid_SPEICHER: process(user_clk)
--    begin
--        if user_clk='1' and user_clk'event then
--            if A15 = '1' then
--                countvalid <= x"00";
--            elsif countvalid = x"FE" then   
--                countvalid <= x"09";
--            else
--                countvalid <= countvalid + 1;
--            end if;
--        end if;
--    end process;
--    
--    countvalidmask_SPEICHER: process(user_clk, countvalid, countvalidmask)
--    begin
--        if user_clk = '1' and user_clk'event then
--            if countvalid = x"08" then
--                countvalidmask <= '0';
--            elsif countvalid = x"00" then
--                countvalidmask <= '1';
--            else
--                countvalidmask <= countvalidmask;
--            end if;
--        end if;
--    end process;


    DATA_OUT: process (nRD, A15, Dout, nWR)
    begin
    
        DIR <= not nRD;
        
        nOE <= not ((not nRD and not A15) or ( not nWR and not A15));
        
        if nRD = '0' then -- when READ
            D <= Dout;
        else
            D <= (others => 'Z');

        end if;

    end process;

    Din <= D;

end rtl;
